'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var types = _ref.types;


  /**
   * Used to track variables built during the AST pass. We instantiate these in
   * the `Program` visitor in order to support running the plugin in watch mode
   * or on multiple files.
   *
   * @type Store
   */
  var store = new _Store2.default(_mapping2.default.module.has('fp') ? [_mapping2.default.id, 'lodash/fp'] : [_mapping2.default.id]);

  function buildDeclaratorHandler(key) {
    return function (_ref2) {
      var node = _ref2.node;

      var decNode = node[key];
      if (types.isIdentifier(decNode)) {
        node[key] = store.getValueBy('member', decNode.name) || decNode;
      }
    };
  }

  function buildExpressionHandler(props) {
    return function (_ref3) {
      var node = _ref3.node;

      _lodash2.default.each(props, function (key) {
        var expNode = node[key];
        if (types.isIdentifier(expNode)) {
          node[key] = store.getValueBy('member', expNode.name) || expNode;
        }
      });
    };
  }

  function getImportBase(pkgStore) {
    return _lodash2.default.get(pkgStore, 'id') == 'lodash/fp' ? 'fp' : '';
  }

  function isDefaultImport(name) {
    return !!store.getStoreBy('default', name);
  }

  /*--------------------------------------------------------------------------*/

  var visitor = {

    'Program': {
      enter: function enter(path, state) {
        var _$assign = _lodash2.default.assign(_mapping2.default, (0, _config2.default)(state.opts));

        var id = _$assign.id;

        if (!id) {
          throw new Error('Cannot find Lodash module');
        }
        if (!store.has(id)) {
          store.set(id);
        }
        // Clear tracked Lodash method imports and variables.
        _importModule2.default.cache.clear();
        store.clear();
      }
    },

    ImportDeclaration: function ImportDeclaration(path) {
      var file = path.hub.file;
      var node = path.node;
      var pkgId = node.source.value;

      var pkgStore = store.get(pkgId);

      if (!pkgStore) {
        return;
      }
      var importBase = getImportBase(pkgStore);
      var defaultMap = pkgStore.get('default');
      var memberMap = pkgStore.get('member');

      // Remove the original import node to be replaced.
      path.remove();

      // Track all the Lodash default and specifier imports in the source.
      _lodash2.default.each(node.specifiers, function (spec) {
        var localName = spec.local.name;
        if (types.isImportSpecifier(spec)) {
          // Handle import specifier (i.e. `import {map} from 'lodash'`).
          var identifier = (0, _importModule2.default)(spec.imported.name, file, importBase, localName);
          memberMap.set(localName, identifier);
        } else {
          // Handle default specifier (i.e. `import _ from 'lodash'`).
          defaultMap.set(localName, true);
        }
      });
    },
    MemberExpression: function MemberExpression(path) {
      var file = path.hub.file;
      var node = path.node;
      var object = node.object;


      if (!types.isIdentifier(object)) {
        return;
      }
      var pkgStore = store.getStoreBy('default', object.name);
      if (pkgStore) {
        var key = node.property.name;
        if (key == 'chain') {
          throw new Error(CHAIN_ERROR);
        }
        // Transform `_.foo` to `_foo`.
        path.replaceWith((0, _importModule2.default)(key, file, getImportBase(pkgStore)));
      } else {
        // Allow things like `bind.placeholder = {}`.
        node.object = store.getValueBy('member', object.name) || object;
      }
    },
    CallExpression: function CallExpression(path) {
      var file = path.hub.file;
      var node = path.node;
      var callee = node.callee;


      if (types.isIdentifier(callee)) {
        var name = callee.name;

        if (isDefaultImport(name)) {
          // Detect chain sequences via _(value).
          throw new Error(CHAIN_ERROR);
        }
        // Update the import specifier if it's marked for replacement.
        node.callee = store.getValueBy('member', name) || callee;
      } else if (types.isMemberExpression(callee)) {
        visitor.MemberExpression(path.get('callee'));
      }
      // Support lodash methods as parameters (#11), e.g. `_.flow(_.map, _.head)`.
      _lodash2.default.each(node.arguments, function (arg, index, args) {
        if (types.isIdentifier(arg)) {
          // Assume default imports are placeholders (#33).
          args[index] = isDefaultImport(arg.name) ? types.memberExpression(node.callee, types.identifier('placeholder')) : store.getValueBy('member', arg.name) || arg;
        }
      });
    },
    ExportNamedDeclaration: function ExportNamedDeclaration(path) {
      var node = path.node;
      var file = path.hub.file;

      var pkgId = _lodash2.default.get(node, 'source.value');
      var pkgStore = store.get(pkgId);
      var importBase = getImportBase(pkgStore);

      if (pkgStore) {
        node.source = null;
      }
      _lodash2.default.each(node.specifiers, function (spec) {
        var localName = spec.local.name;
        spec.local = pkgStore ? (0, _importModule2.default)(localName, file, importBase) : store.getValueBy('member', localName) || spec.local;
      });
    },


    // Various other (less common) ways to use a lodash specifier. This code
    // doesn't apply to uses on a lodash object only directly imported specifiers.

    // See #34.
    'Property': buildDeclaratorHandler('value'),
    'VariableDeclarator': buildDeclaratorHandler('init'),

    // Allow things like `o.a = _.noop`.
    'AssignmentExpression': buildExpressionHandler(['right']),

    // Allow things like `var x = y || _.noop`. See #28.
    'LogicalExpression': buildExpressionHandler(['left', 'right']),

    // Allow things like `var x = y ? _.identity : _.noop`. See #28.
    'ConditionalExpression': buildExpressionHandler(['test', 'consequent', 'alternate'])
  };

  return { visitor: visitor };
};

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

var _importModule = require('./importModule');

var _importModule2 = _interopRequireDefault(_importModule);

var _mapping = require('./mapping');

var _mapping2 = _interopRequireDefault(_mapping);

var _Store = require('./Store');

var _Store2 = _interopRequireDefault(_Store);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/** The error message used when chain sequences are detected. */
var CHAIN_ERROR = ['Lodash chain sequences are not supported by babel-plugin-lodash.', 'Consider substituting chain sequences with _.flow composition patterns.', 'See https://medium.com/making-internets/why-using-chain-is-a-mistake-9bc1f80d51ba'].join('\n');

/*----------------------------------------------------------------------------*/

;
module.exports = exports['default'];